import datetime
from typing import (
    TYPE_CHECKING,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    cast,
)

from .exceptions import PrepareError
from .structure import FinId, Mic, OlsonTimezone, Weekday, WeekdayPeriod, WeekdaySet
from .util import snake_case, snake_dict

if TYPE_CHECKING:
    from .catalog import Catalog

T = TypeVar("T")


class BaseObject:
    """Base model objects"""

    def __init__(self, data: Dict):
        self.data = data

    def to_dict(self) -> Dict:
        return self.data

    def to_tuple(self) -> Tuple:
        all_values = []
        for current_field in self.fields:
            current_value = getattr(self, current_field.field_name)
            all_values.append(current_value)
        return tuple(all_values)

    @classmethod
    def from_tuple(cls, data: Tuple):
        data_dict = {}
        for index, current_field in enumerate(cls.fields):
            current_value = data[index]
            data_dict[current_field.field_name] = current_value
        return cls(data_dict)

    @classmethod
    def from_dict(cls, data: Dict):
        normalized = snake_dict(data)
        return cls(normalized)

    @classmethod
    def get_catalog(cls, catalog: Optional["Catalog"]) -> "Catalog":
        if catalog is None:
            from .catalog import default_catalog

            return default_catalog
        return catalog

    def __repr__(self):
        class_name = self.__class__.__name__
        return f"{class_name}({self.data!r})"

    def __str__(self):
        class_name = self.__class__.__name__
        all_str = []
        for current_field in self.fields:
            current_value = getattr(self, current_field.field_name)
            if current_value:
                all_str.append(str(current_value))
        fields_str = " ".join(all_str)
        return f"{class_name} {fields_str}"


class Field(Generic[T]):
    """Base field class"""

    def __set_name__(self, owner: Type[BaseObject], name):
        if not hasattr(owner, "fields"):
            owner.fields: List["Field"] = []
        owner.fields.append(self)
        self.field_name = name

    def __get__(self, obj, objtype=None) -> T:
        if obj is None:
            return self
        key = self.field_name
        if key in obj.data:
            value = obj.data[key]
        else:
            key = snake_case(objtype.__name__) + "_" + key
            value = obj.data.get(key, None)
        if value is None or value == "":
            return None
        else:
            return self.safe_prepare(value)

    def safe_prepare(self, value: str) -> T:
        try:
            return self.prepare(value)
        except Exception as error:
            raise PrepareError(self, value, inner=error)

    def prepare(self, value: str) -> T:
        return cast(T, value)


class StringField(Field[str]):
    """Field of string type"""

    pass


class BooleanField(Field[bool]):
    """Field of boolean type"""

    pass


class IntegerField(Field[int]):
    """Field of int type"""

    def prepare(self, value: str) -> int:
        return int(value)


class DateField(Field[datetime.date]):
    """Field of date type"""

    def prepare(self, value: str) -> datetime.date:
        return datetime.date.fromisoformat(value)


class DateTimeField(Field[datetime.datetime]):
    """Field of datetime type"""

    def prepare(self, value: str) -> datetime.datetime:
        return datetime.datetime.fromisoformat(value)


class TimeField(Field[datetime.time]):
    """Field of time type"""

    def prepare(self, value: str) -> datetime.time:
        return datetime.time.fromisoformat(value)


class ReferenceField(Field[BaseObject]):
    """Field for referencing other BaseObject children"""

    def __init__(self, referenced_type):
        super().__init__()
        self.referenced_type = referenced_type


class ListField(Field[List[T]]):
    """Field of a list with specific type"""

    pass


class OlsonTimezoneField(Field[OlsonTimezone]):
    """Field of an Olson Timezone"""

    def prepare(self, value: str) -> OlsonTimezone:
        return OlsonTimezone.from_string(value)


class WeekdayField(Field[Weekday]):
    """Field for a Weekday"""

    def prepare(self, value: str) -> Weekday:
        return Weekday.from_string(value)


class WeekdayPeriodField(Field[WeekdayPeriod]):
    """Field for period like Mon-Fri"""

    def prepare(self, value: str) -> Weekday:
        return WeekdayPeriod.from_string(value)


class WeekdaySetField(Field[WeekdaySet]):
    """Field for set of periods like Mon-Fri"""

    def prepare(self, value: str) -> WeekdaySet:
        return WeekdaySet.from_string(value)


class FinIdField(Field[FinId]):
    """Field for a FinID"""

    def prepare(self, value: str) -> FinId:
        return FinId.from_string(value)


class MicField(Field[Mic]):
    """Field for a MIC"""

    def prepare(self, value: str) -> Mic:
        return Mic.from_string(value)



import csv
import json
import re
from io import StringIO
from typing import Dict

from .validate import validate_instance_arg, validate_str_arg


def snake_case(text):
    text = validate_str_arg("text", text, strip=True)
    if "-" in text:
        text = text.replace("-", "_")
    words = re.findall(r"[A-Z]?[a-z]+|[A-Z]+|[a-z]+", text)
    return "_".join(words).lower()


def snake_dict(data: dict):
    data = validate_instance_arg("data", data, dict)
    snake_data = {}
    for key, value in data.items():
        snake_key = snake_case(key)
        snake_data[snake_key] = value
    return snake_data


def slugify(name):
    name = validate_str_arg("name", name, strip=True)
    slug = []
    just_added_dash = False

    for char in name.lower():
        if char.isalnum():  # Check if the character is alphanumeric
            slug.append(char)
            just_added_dash = False
        elif not just_added_dash:
            slug.append("-")
            just_added_dash = True

    return "".join(slug)


def create_table_structure(data, num_columns=3, suppress_empty_values=False):
    sorted_keys = sorted(data.keys())

    # Filter out empty values if suppress_empty_values is True
    if suppress_empty_values:
        sorted_keys = [key for key in sorted_keys if data[key] != ""]

    # Calculate the number of rows needed
    num_rows = len(sorted_keys) // num_columns
    if len(sorted_keys) % num_columns != 0:
        num_rows += 1

    # Initialize an empty table structure
    table = []

    for i in range(num_columns):
        column_keys = sorted_keys[i * num_rows : (i + 1) * num_rows]
        max_key_width = max(len(key) for key in column_keys)
        max_value_width = max(len(str(data[key])) for key in column_keys)

        column_data = [(key, data[key]) for key in column_keys]
        column_metadata = (max_key_width, max_value_width)

        table.append((column_metadata, column_data))

    return table


def render_ascii_table(table_structure, fixed_value_width=None, max_width=None):
    num_columns = len(table_structure)
    num_rows = max(len(column_data) for _, column_data in table_structure)

    table = ""

    for row_index in range(num_rows):
        for col_index in range(num_columns):
            metadata, rows = table_structure[col_index]
            max_key_width, max_value_width = metadata
            if row_index < len(rows):
                key, value = rows[row_index]
            else:
                key = ""
                value = ""

            # Use the fixed_value_width if provided, else use the max_value_width
            value_width = (
                fixed_value_width if fixed_value_width is not None else max_value_width
            )
            if max_width:
                value_width = min(max_width, value_width)

            # Truncate long string values to fit the width
            if len(value) > value_width:
                value = value[:value_width]

            # Calculate the number of dots to ensure at least one dot
            dots = "." * (max_key_width - len(key) + 1)

            if key == "" and value == "":
                key_width = len(key) + len(dots)
                table += " " * (key_width + value_width + 2)
            else:
                table += f"{key}{dots}: {value:{value_width}}"

            if col_index < num_columns - 1:
                table += "  "  # Add spacing between columns
            else:
                table += "\n"  # Newline at the end of the row

    return table


def pprint_data(data: Dict):
    table = create_table_structure(data, suppress_empty_values=True)
    ascii = render_ascii_table(table, max_width=15)
    print(ascii)


def get_csv_from_tuple(data: tuple) -> str:
    csv_data = StringIO()
    writer = csv.writer(csv_data)
    writer.writerow(data)
    csv_string = csv_data.getvalue()
    csv_data.close()
    return csv_string


def get_tuple_from_csv(csv_string: str) -> tuple:
    csv_data = StringIO(csv_string)
    reader = csv.reader(csv_data)
    row = next(reader)
    csv_data.close()
    return tuple(row)


class StrEncoder(json.JSONEncoder):
    """Helps encoding flat data"""

    def default(self, obj):
        try:
            encoded = super().default(obj)
        except TypeError:
            encoded = str(obj)
        return encoded



import datetime
import json
import shutil
import tempfile
import zipfile
from contextlib import contextmanager
from functools import cached_property
from pathlib import Path
from typing import Optional
from urllib.error import HTTPError
from urllib.parse import urljoin
from urllib.request import Request, urlopen

from .config import main_config
from .exceptions import ClientError, TokenError
from .typing import StrOrPath
from .validate import validate_instance_arg, validate_path_arg


class Client:
    """TradingHours API client"""

    def __init__(self, token, base_url):
        self.token = token
        self.base_url = base_url

    @contextmanager
    def get_response(self, path):
        url = urljoin(self.base_url, path)
        request = Request(url)
        request.add_header("Authorization", f"Bearer {self.token}")
        try:
            yield urlopen(request)
        except HTTPError as error:
            if error.code == 401:
                raise TokenError("Token is missing or invalid")
            raise ClientError("Error getting server response", inner=error)

    @contextmanager
    def download_temporary(self, path):
        with self.get_response(path) as response:
            with tempfile.NamedTemporaryFile() as temp_file:
                shutil.copyfileobj(response, temp_file)
                temp_file.flush()
                temp_file.seek(0)
                yield temp_file

    def get_json(self, path):
        with self.get_response(path) as response:
            data = json.load(response)
        return data


class DataManager:
    """Manages accessing remote data"""

    def __init__(self, client: Client, root: StrOrPath):
        self.client = validate_instance_arg("client", client, Client)
        self._root = validate_path_arg("root", root)

    @property
    def root(self) -> Path:
        return self._root

    @property
    def csv_dir(self) -> Path:
        return self.root / "csv"

    @cached_property
    def remote_timestamp(self) -> datetime.datetime:
        data = default_client.get_json("last-updated")
        last_updated = data["last_updated"]
        timestamp = datetime.datetime.fromisoformat(last_updated)
        return timestamp

    @cached_property
    def local_timestamp(self) -> Optional[datetime.datetime]:
        version_file = self.root / "VERSION.txt"
        if version_file.exists():
            timestamp_format = "Generated at %a, %d %b %Y %H:%M:%S %z"
            content = version_file.read_text()
            line = content.splitlines()[0]
            timestamp = datetime.datetime.strptime(line, timestamp_format)
            return timestamp
        return None

    @cached_property
    def needs_download(self) -> bool:
        if self.local_timestamp:
            return self.remote_timestamp > self.local_timestamp
        return True

    def download(self):
        self.root.mkdir(parents=True, exist_ok=True)
        with default_client.download_temporary("download") as temp_file:
            with zipfile.ZipFile(temp_file, "r") as zip_ref:
                zip_ref.extractall(self.root)


default_client = Client(
    main_config.get("api", "token"),
    main_config.get("api", "base_url"),
)

default_data_manager = DataManager(
    default_client,
    main_config.get("data", "remote_dir"),
)



import os
from typing import Optional, Tuple

from tradinghours.config import main_config
from tradinghours.store.base import Collection, Registry


class Store:
    """Manages data loaded into memory"""

    def __init__(self, collections_registry: Registry):
        self._collections = collections_registry

    @property
    def collections(self) -> Registry[Collection]:
        return self._collections

    @property
    def token(self):
        return os.getenv("TRADINGHOURS_TOKEN")

    def clear_collection(self, name: str):
        collection_obj = self.collections.get(name)
        collection_obj.clear()

    def store_tuple(
        self,
        data: Tuple,
        collection,
        cluster: Optional[str] = None,
        key: Optional[str] = None,
    ):
        collection_obj = self.collections.get(collection)
        if cluster is None:
            cluster = "default"
        cluster_obj = collection_obj.clusters.get(cluster)
        cluster_obj.append(key, data)

    def flush(self):
        for collection in self.collections:
            for cluster in collection.clusters:
                cluster.flush()


def create_file_store():
    from tradinghours.store.file import FileCollectionRegistry

    root = main_config.get("data", "local_dir")
    registry = FileCollectionRegistry(root)
    store = Store(registry)
    return store


def create_sql_store():
    from tradinghours.store.sql import SqlCollectionRegistry

    db_url = main_config.get("data", "db_url")
    registry = SqlCollectionRegistry(db_url)
    store = Store(registry)
    return store


def create_default_store():
    if main_config.getboolean("data", "use_db"):
        return create_sql_store()
    else:
        return create_file_store()


default_store = create_default_store()



import datetime
from typing import List

from tradinghours.season import SeasonDefinition

from .base import (
    BaseObject,
    BooleanField,
    DateField,
    DateTimeField,
    FinIdField,
    IntegerField,
    ListField,
    OlsonTimezoneField,
    StringField,
    TimeField,
    WeekdayField,
    WeekdaySetField,
)
from .typing import StrOrDate, StrOrFinId
from .validate import validate_date_arg, validate_finid_arg, validate_range_args


class ConcretePhase(BaseObject):
    """A period within a schedule"""

    phase_type = StringField()
    """Well known options"""

    phase_name = StringField()
    """Describes the name for the phase type."""

    phase_memo = StringField()
    """If applicable, will have additional description or information."""

    status = StringField()
    """Describes what status the market is currently."""

    start = DateTimeField()
    """The date the market phase type started."""

    end = DateTimeField()
    """The scheduled date for the market phase type to end."""


class DateSchedule(BaseObject):
    """Full trading schedule for a market on a specific date"""

    date = DateField()
    """The date for the data returned."""

    day_of_week = WeekdayField()
    """The day of the week for the data returned."""

    is_open = BooleanField()
    """Describes in true/false statement if the market is open."""

    has_settlement = BooleanField()
    """Describes in true/false statement if the market has settlement."""

    holiday = StringField()
    """Describes the holiday, if any."""

    schedule = ListField[ConcretePhase]()
    """Nested data of the schedule."""


class PeriodSchedule(BaseObject):
    """Market phases for a given period"""

    start = DateTimeField()
    """The start date for the data returned."""

    end = DateTimeField()
    """The end date for the data returned."""

    schedule = ListField[ConcretePhase]()
    """Nested data of the schedule."""


class Schedule(BaseObject):
    """Schedules definitions from TradingHours"""

    fin_id = FinIdField()
    schedule_group = StringField()
    schedule_group_memo = StringField()
    timezone = OlsonTimezoneField()
    phase_type = StringField()
    phase_name = StringField()
    phase_memo = StringField()
    days = WeekdaySetField()
    start = TimeField()
    end = TimeField()
    offset_days = IntegerField()
    duration = StringField()
    min_start = TimeField()
    max_start = TimeField()
    min_end = TimeField()
    max_end = TimeField()
    in_force_start_date = DateField()
    in_force_end_date = DateField()
    season_start = StringField()
    season_end = StringField()

    @property
    def has_season(self) -> bool:
        season_start = (self.season_start or "").strip()
        season_end = (self.season_end or "").strip()
        return season_start and season_end

    def is_in_force(self, start: StrOrDate, end: StrOrDate) -> bool:
        start, end = validate_range_args(
            validate_date_arg("start", start),
            validate_date_arg("end", end),
        )
        if self.in_force_start_date is None and self.in_force_end_date is None:
            return True
        elif self.in_force_start_date is None:
            return self.in_force_end_date >= start
        elif self.in_force_end_date is None:
            return self.in_force_start_date <= end
        else:
            return self.in_force_start_date <= end and self.in_force_end_date >= start

    def match_occurrences(self, some_date: StrOrDate) -> List[datetime.date]:
        """This method will return all matches for one single date"""
        some_date = validate_date_arg("some_date", some_date)

        # Keep track of all dates matching some_date, considering the offset
        # for previous dates could match this date too
        occurrences: List[datetime.date] = []

        # We will scan all dates considering the offset
        current_date = some_date
        current_offset = self.offset_days
        while current_offset >= 0:
            # Check whether it happens on this specific date
            happens = self.is_in_force(current_date, current_date)
            happens = happens and self.days.matches(current_date)
            if happens:
                occurrences.append(current_date)

            # Prepare to match now the previous date
            current_date -= datetime.timedelta(days=1)
            current_offset -= 1

        # Return all occurences
        return occurrences

    def match_season(self, some_date: StrOrDate) -> bool:
        """Indicates whether some_date matches season if any"""
        some_date = validate_date_arg("some_date", some_date)

        # If there is no season, it means there is no restriction in terms
        # of the season when this schedule is valid, and as such it is valid,
        # from a season-perspective for any date
        if not self.has_season:
            return True

        start_date = SeasonDefinition.get_date(self.season_start, some_date.year)
        end_date = SeasonDefinition.get_date(self.season_end, some_date.year)
        if end_date < start_date:
            return some_date <= end_date or some_date >= start_date
        return some_date >= start_date and some_date <= end_date

    @classmethod
    def list_all(cls, finid: StrOrFinId, catalog=None) -> List["Schedule"]:
        finid = validate_finid_arg("finid", finid)
        catalog = cls.get_catalog(catalog)
        return list(map(lambda t: t[1], catalog.list(Schedule, cluster=str(finid))))

    @classmethod
    def is_group_open(cls, group):
        # TODO: implement a ScheduleGroup type and consider other open groups
        return group.lower() == "regular"


class RegularSchedule(BaseObject):
    """Repeating schedule for a market"""

    day = WeekdayField()
    """Day of the week in string format."""

    open = BooleanField()
    """Describes if the market is open in true/false."""

    time_start = TimeField()
    """Describes the time the market trading session opens."""

    time_end = TimeField()
    """Describes the time the market trading session ends."""

    lunch = BooleanField()
    """Describes if the market has observed lunch hours in true/false."""

    lunch_start = TimeField()
    """If observed lunch hours, this describes when lunch hours start."""

    lunch_end = TimeField()
    """If observing lunch hours, this describes when lunch hours end."""

    pre_hours_start = TimeField()
    """If pre-hours, describes what time they start."""

    pre_hours_end = TimeField()
    """If pre-hours, describes what time they end."""

    post_hours_start = TimeField()
    """If post-hours, describes what time they start."""

    post_hours_end = TimeField()
    """If post-hours, describes what time they end."""



from typing import Dict, Iterator, Tuple

from tradinghours.exceptions import MissingSqlAlchemyError

try:
    from sqlalchemy import (
        JSON,
        Column,
        Engine,
        Integer,
        MetaData,
        String,
        Table,
        create_engine,
        inspect,
    )
except ImportError:
    raise MissingSqlAlchemyError("SQLAlchemy not installed")


from tradinghours.store.base import Cluster, Collection, Registry
from tradinghours.util import get_csv_from_tuple, get_tuple_from_csv

TABLE_NAME_PREFIX = "thstore_"


class SqlCluster(Cluster):
    """Manages one page in a SQL database table"""

    DEFAULT_CACHE_SIZE = 500

    def __init__(self, engine: Engine, table: Table, slug: str, cache_size=None):
        self._engine = engine
        self._table = table
        self._slug = slug
        self._cached = []
        self._cache_size = cache_size or self.DEFAULT_CACHE_SIZE

    def truncate(self):
        with self._engine.connect() as connection:
            connection.execute(f"DELETE FROM {self._table.name}")

    def flush(self):
        with self._engine.connect() as connection:
            with connection.begin() as transaction:
                for key, data in self._cached:
                    data_csv = get_csv_from_tuple(data)
                    db_record = {"slug": self._slug, "key": key, "data": data_csv}
                    connection.execute(self._table.insert().values(db_record))
                transaction.commit()
        self._cached = []

    def load_all(self) -> Dict[str, Tuple]:
        keyed_items = {}
        with self._engine.connect() as connection:
            select_query = (
                self._table.select()
                .add_columns(self._table.c.key, self._table.c.data)
                .where(self._table.c.slug == self._slug)
            )
            result = connection.execute(select_query)
            for row in result:
                key = row.key or str(row.id)
                data = get_tuple_from_csv(row.data)
                keyed_items[key] = data
        return keyed_items


class SqlClusterRegistry(Registry[SqlCluster]):
    """Holds a series of SQL clusters"""

    def __init__(self, engine: Engine, table: Table):
        self._engine = engine
        self._table = table
        super().__init__()

    def create(self, slug: str) -> SqlCluster:
        return SqlCluster(self._engine, self._table, slug)

    def discover(self) -> Iterator[str]:
        with self._engine.connect() as connection:
            select_query = (
                self._table.select().add_columns(self._table.c.slug).distinct()
            )
            result = connection.execute(select_query)
            for row in result:
                yield row.slug


class SqlCollection(Collection):
    def __init__(self, engine: Engine, table: Table):
        self._engine = engine
        self._table = table
        self.touch()
        self._clusters = SqlClusterRegistry(self._engine, self._table)

    @property
    def clusters(self) -> SqlClusterRegistry:
        return self._clusters

    def touch(self):
        self._table.create(self._engine, checkfirst=True)

    def clear(self):
        for current in self._clusters:
            current.truncate()
        self._clusters = SqlClusterRegistry(self._engine, self._table)


class SqlCollectionRegistry(Registry[SqlCollection]):
    def __init__(self, db_url: str):
        self._db_url = db_url
        self._engine = create_engine(self._db_url)
        self._metadata = MetaData()
        super().__init__()

    def create(self, slug: str) -> SqlCollection:
        table_name = TABLE_NAME_PREFIX + slug.replace("-", "_")
        table = Table(
            table_name,
            self._metadata,
            Column("id", Integer, primary_key=True),
            Column("slug", String),
            Column("key", String),
            Column("data", JSON),
        )
        collection = SqlCollection(self._engine, table)
        collection.touch()
        return collection

    def discover(self) -> Iterator[str]:
        inspector = inspect(self._engine)
        all_names = inspector.get_table_names()
        for table_name in all_names:
            if table_name.startswith(TABLE_NAME_PREFIX):
                yield table_name[len(TABLE_NAME_PREFIX) :]



import configparser
import os
from pathlib import Path

PROJECT_PATH = Path(__file__).parent

# Define default settings in this dictionary
default_settings = {
    "api": {
        "token": os.getenv("TRADINGHOURS_TOKEN", ""),
        "base_url": "https://api.tradinghours.com/v3/",
    },
    "data": {
        "use_db": False,
        "local_dir": PROJECT_PATH / "store_dir" / "local",
        "remote_dir": PROJECT_PATH / "store_dir" / "remote",
        "db_url": f"sqlite:///{PROJECT_PATH / 'store_dir' / 'local.db'}",
    },
}

# Read config file with defaults
main_config = configparser.ConfigParser()
main_config.read_dict(default_settings)
main_config.read("tradinghours.ini")



from functools import cached_property
from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    from .base import Field


class TradingHoursError(Exception):
    """Baseclass for all errors from this library"""

    def __init__(self, message: str, inner: Optional[Exception] = None):
        super().__init__(message, inner)
        self._message = message
        self._inner = inner

    @property
    def message(self):
        return self._message

    @property
    def inner(self):
        return self._inner

    @cached_property
    def detail(self):
        return self.build_detail()

    @cached_property
    def help_message(self):
        return self.build_help_message()

    def build_detail(self, message: Optional[str] = None):
        message = message or self.message
        if self.inner:
            message = f"{message} ({self.inner})"
        return message

    def build_help_message(self):
        return None

    def __str__(self):
        return self.detail


class PrepareError(TradingHoursError):
    """Happens when a field from a model cannot be interpreted"""

    def __init__(self, field: "Field", value: Any, inner: Optional[Exception] = None):
        super().__init__("Error preparing field", inner=inner)
        self._field = field
        self._value = value

    @property
    def field(self):
        return self._field

    @property
    def value(self):
        return self._value

    def build_detail(self):
        message = f"Could not prepare field {self._field.field_name}"
        return super().build_detail(message)


class ClientError(TradingHoursError):
    """When an error occurs accessing remote HTTP server"""

    pass


class TokenError(ClientError):
    """When server access fails because of an invalid token"""

    def build_help_message(self):
        return (
            "A TradingHours token is required to perform this operation. "
            "You can access https://www.tradinghours.com/user/api-tokens to "
            "obtain one. In case you already have a token, remember to make "
            "it available by exporting the environment variable "
            "TRADINGHOURS_TOKEN and try again."
        )


class MissingDefinitionError(TradingHoursError):
    """When a season definition is not found"""

    pass


class MissingSqlAlchemyError(TradingHoursError):
    """When SQL Alchemy is not installed"""

    def build_help_message(self):
        return (
            "You need to install SQLAlchemy in order to usee database "
            "ready store. You should be able to do that by running "
            "`pip install tradinghours[sql]` from the command line."
        )



from abc import ABC, abstractmethod, abstractproperty
from typing import Dict, Generic, Iterator, List, Optional, Tuple, TypeVar

from ..util import slugify

T = TypeVar("T")


class Registry(ABC, Generic[T]):
    """Keeps track of keyed resources"""

    def __init__(self):
        self._resources = {}
        for name in self.discover():
            self.get(name)

    def get(self, name: str) -> T:
        slug = slugify(name)
        resource = self._resources.get(slug, None)
        if resource is None:
            resource = self.create(slug)
            self._resources[slug] = resource
        return resource

    @abstractmethod
    def create(self, slug: str) -> T:
        raise NotImplementedError()

    @abstractmethod
    def discover(self) -> Iterator[str]:
        raise NotImplementedError()

    def __iter__(self) -> Iterator[T]:
        return iter(self._resources.values())


class Cluster(ABC):
    """Generic Cluster interface"""

    def __init__(self, cache_size: Optional[int] = None):
        self._cached: List[str, Tuple] = []
        self._cache_size = cache_size or self.DEFAULT_CACHE_SIZE

    @abstractmethod
    def truncate(self):
        """Clear all elements from this cluster"""
        raise NotImplementedError()

    def append(self, key: Optional[str], data: Tuple):
        """Appends one element to this cluster

        Calling this method does not indicate that the data will be saved in
        the underlying storage. You need to call `flush` for that.

        """
        record = [key, data]
        self._cached.append(record)
        if len(self._cached) >= self._cache_size:
            self.flush()

    @abstractmethod
    def flush(self):
        """Flushes data to the underlying storage"""
        raise NotImplementedError()

    @abstractmethod
    def load_all(self) -> Dict[str, Tuple]:
        """Loads all data from this cluster"""
        raise NotImplementedError()


class Collection(ABC):
    """Generic Collection interface"""

    @abstractproperty
    def clusters(self) -> Registry:
        """Gets the clusters registry for this collection"""
        raise NotImplementedError()

    @abstractmethod
    def touch(self):
        """Ensure the collection exists"""
        raise NotImplementedError()

    @abstractmethod
    def clear(self):
        """Wipes all clusters in this collection"""
        raise NotImplementedError()



import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from .structure import FinId, Weekday


StrOrPath = Union[str, Path]
StrOrDate = Union[str, datetime.date]
StrOrFinId = Union[str, "FinId"]
WeekdaySpec = Union[str, int, datetime.date, datetime.datetime, "Weekday"]



import csv
from pathlib import Path
from typing import Generator, Generic, Type, TypeVar

from tradinghours.base import BaseObject
from tradinghours.typing import StrOrPath
from tradinghours.validate import (
    validate_path_arg,
    validate_str_arg,
    validate_subclass_arg,
)

B = TypeVar("B", bound=BaseObject)


class SourceFile(Generic[B]):
    """Represents a file to be imported"""

    def __init__(self, root: StrOrPath, name: str, model: Type[B]):
        self._name = validate_str_arg("name", name)
        self._root = validate_path_arg("root", root)
        self._model = validate_subclass_arg("model", model, BaseObject)

    @property
    def name(self) -> str:
        return self._name.replace("-", "_")

    @property
    def root(self) -> Path:
        return self._root

    @property
    def model(self) -> Type[B]:
        return self._model

    @property
    def filename(self) -> str:
        return self._name.replace("_", "-") + ".csv"

    @property
    def path(self) -> Path:
        return self._root / self.filename

    def load_iter(self) -> Generator[B, None, None]:
        with open(self.path, "r", encoding="utf-8-sig", errors="replace") as file:
            for row in csv.DictReader(file):
                yield self.model.from_dict(row)



import datetime
from datetime import timedelta
from typing import Dict, Generator, Iterable, List, Tuple

from .base import (
    BaseObject,
    BooleanField,
    DateField,
    FinIdField,
    MicField,
    StringField,
    WeekdaySetField,
)
from .schedule import ConcretePhase, Schedule
from .typing import StrOrDate, StrOrFinId
from .validate import (
    validate_date_arg,
    validate_finid_arg,
    validate_range_args,
    validate_str_arg,
)

# Arbitrary max offset days for TradingHours data
MAX_OFFSET_DAYS = 2


class Market(BaseObject):
    """One known market for TradingHours"""

    fin_id = FinIdField()
    """The FinID for the market."""

    exchange_name = StringField()
    """The exchange name of the market."""

    market_name = StringField()
    """The name of the market."""

    security_group = StringField()
    """Description of the products or securities group."""

    mic = MicField()
    """The MIC for the market."""

    acronym = StringField()
    """The acronym for the market."""

    asset_type = StringField()
    """Describes the asset type of the market."""

    memo = StringField()
    """A description or additional details about the trading venue."""

    permanently_closed = DateField()
    """If a market is permanently closed, this shows the date."""

    timezone = StringField()
    """Gives the timezone the market utilizes."""

    weekend_definition = WeekdaySetField()
    """Indicates the days of the week when the market regularly closed."""

    replaced_by = FinIdField()

    @property
    def country_code(self):
        """Two-letter country code."""
        return self.fin_id.country

    def _pick_schedule_group(
        self,
        some_date: datetime.date,
        holidays: Dict[datetime.date, "MarketHoliday"],
    ) -> Tuple[str, bool]:
        if found := holidays.get(some_date):
            schedule_group = found.schedule.lower()
            fallback = Schedule.is_group_open(schedule_group)
        else:
            schedule_group = "regular"
            fallback = False
        return schedule_group, fallback

    def _filter_schedule_group(
        self, schedule_group: str, schedules: Iterable[Schedule]
    ) -> Iterable[Schedule]:
        for current in schedules:
            if current.schedule_group.lower() == schedule_group:
                yield current

    def _filter_inforce(
        self, some_date: datetime.date, schedules: Iterable[Schedule]
    ) -> Iterable[Schedule]:
        for current in schedules:
            if current.is_in_force(some_date, some_date):
                yield current

    def _filter_season(
        self, some_date: datetime.date, schedules: Iterable[Schedule]
    ) -> Iterable[Schedule]:
        for current in schedules:
            if current.match_season(some_date):
                yield current

    def _filter_weekdays(
        self, some_date: datetime.date, schedules: Iterable[Schedule]
    ) -> Iterable[Schedule]:
        for current in schedules:
            if current.days.matches(some_date):
                yield current

    def generate_schedules(
        self, start: StrOrDate, end: StrOrDate, catalog=None
    ) -> Generator[ConcretePhase, None, None]:
        start, end = validate_range_args(
            validate_date_arg("start", start),
            validate_date_arg("end", end),
        )
        catalog = self.get_catalog(catalog)

        # Get required global data
        offset_start = start - timedelta(days=MAX_OFFSET_DAYS)
        all_schedules = Schedule.list_all(self.fin_id)
        holidays = MarketHoliday.build_keyed(self.fin_id, offset_start, end)

        # Iterate through all dates generating phases
        current_date = offset_start
        while current_date <= end:
            # Starts with all schedules
            schedules = all_schedules

            # Filter schedule group based on holiday if any
            schedule_group, fallback = self._pick_schedule_group(current_date, holidays)
            schedules = self._filter_schedule_group(schedule_group, schedules)

            # Filters what is in force or for expected season
            schedules = self._filter_inforce(current_date, schedules)
            schedules = self._filter_season(current_date, schedules)

            # Save for fallback and filter weekdays
            before_weekdays = list(schedules)
            found_schedules = list(self._filter_weekdays(current_date, before_weekdays))

            # Consider fallback if needed
            if not found_schedules and fallback:
                initial_weekday = current_date.weekday()
                fallback_weekday = 6 if initial_weekday == 0 else initial_weekday - 1
                fallback_schedules = []
                while not fallback_schedules and fallback_weekday != initial_weekday:
                    fallback_schedules = list(
                        filter(
                            lambda s: s.days.matches(fallback_weekday),
                            before_weekdays,
                        ),
                    )
                    fallback_weekday = (
                        6 if fallback_weekday == 0 else fallback_weekday - 1
                    )
                found_schedules = fallback_schedules

            # Sort based on start time and duration
            found_schedules = sorted(
                found_schedules,
                key=lambda s: (s.start, s.duration),
            )

            # Generate phases for current date
            for current_schedule in found_schedules:
                start_date = current_date
                end_date = current_date + timedelta(days=current_schedule.offset_days)

                # Filter out phases not finishing after start because we
                # began looking a few days ago to cover offset days
                if end_date >= start:
                    current_schedule.timezone
                    start_datetime = datetime.datetime.combine(
                        start_date, current_schedule.start
                    )
                    end_datetime = datetime.datetime.combine(
                        end_date, current_schedule.end
                    )
                    start_datetime = current_schedule.timezone.localize(start_datetime)
                    end_datetime = current_schedule.timezone.localize(end_datetime)
                    yield ConcretePhase(
                        dict(
                            phase_type=current_schedule.phase_type,
                            phase_name=current_schedule.phase_name,
                            phase_memo=current_schedule.phase_memo,
                            start=start_datetime.isoformat(),
                            end=end_datetime.isoformat(),
                        )
                    )

            # Next date, please
            current_date += timedelta(days=1)

    @classmethod
    def list_all(cls, catalog=None) -> List["Market"]:
        catalog = cls.get_catalog(catalog)
        return list(catalog.list_all(Market))

    @classmethod
    def get_by_finid(cls, finid: StrOrFinId, follow=True, catalog=None) -> "Market":
        finid = validate_finid_arg("finid", finid)
        catalog = cls.get_catalog(catalog)
        found = catalog.get(Market, str(finid), cluster=finid.country)
        if found and found.replaced_by and follow:
            found = catalog.get(Market, str(found.replaced_by), cluster=finid.country)
        return found

    @classmethod
    def get_by_mic(cls, mic: str, follow=True, catalog=None) -> "Market":
        mic = validate_str_arg("mic", mic)
        catalog = cls.get_catalog(catalog)
        mapping = MicMapping.get(mic, catalog=catalog)
        if mapping:
            return cls.get_by_finid(mapping.fin_id, follow=follow)
        return None

    @classmethod
    def get(cls, identifier: str, follow=True, catalog=None) -> "Market":
        identifier = validate_str_arg("identifier", identifier)
        catalog = cls.get_catalog(catalog)
        if "." in identifier:
            found = cls.get_by_finid(identifier, follow=follow)
        else:
            found = cls.get_by_mic(identifier, follow=follow)
        return found


class MarketHoliday(BaseObject):
    """Holidays for a Market"""

    fin_id = FinIdField()
    """The FinID for the market."""

    date = DateField()
    """Shows the date of the holiday for the market."""

    name = StringField()
    """Describes the name of the holiday."""

    schedule = StringField()
    """Describes if the market closes for the holiday."""

    is_open = BooleanField()
    """Displays in true/false if the market is open for the holiday."""

    has_settlement = BooleanField()
    """Displays in true/false if the market has settlement for the holiday."""

    observed = BooleanField()
    """Displays in true/false if the holiday is observed."""

    memo = StringField()
    """A description or additional details about the holiday."""

    @classmethod
    def list_range(
        cls, finid: StrOrFinId, start: StrOrDate, end: StrOrDate, catalog=None
    ) -> List["MarketHoliday"]:
        finid = validate_finid_arg("finid", finid)
        start, end = validate_range_args(
            validate_date_arg("start", start),
            validate_date_arg("end", end),
        )
        catalog = cls.get_catalog(catalog)
        holidays = list(
            catalog.filter(
                MarketHoliday,
                start.isoformat(),
                end.isoformat(),
                cluster=str(finid),
            )
        )
        return holidays

    @classmethod
    def build_keyed(
        cls, finid: StrOrFinId, start: StrOrDate, end: StrOrDate, catalog=None
    ) -> Dict[datetime.date, "MarketHoliday"]:
        holidays = cls.list_range(finid, start, end)
        keyed = {}
        for current in holidays:
            keyed[current.date] = current
        return keyed


class MicMapping(BaseObject):
    """Mapping from MIC to FinId"""

    mic = StringField()
    """Market Identification Code"""

    fin_id = FinIdField()
    """TradingHours FinId"""

    @classmethod
    def get(cls, mic: str, catalog=None) -> "MicMapping":
        mic = validate_str_arg("mic", mic)
        catalog = cls.get_catalog(catalog)
        return catalog.get(cls, mic)



import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Tuple, Type, TypeVar

if TYPE_CHECKING:
    from .structure import FinId, Weekday

T = TypeVar("T")


def validate_date_arg(name: str, value: Any) -> datetime.date:
    if value is None:
        raise ValueError(f"Missing {name}")
    if isinstance(value, str):
        value = datetime.date.fromisoformat(value)
    if not isinstance(value, datetime.date):
        raise TypeError(f"Invalid {name} type")
    return value


def validate_range_args(start: T, end: T) -> Tuple[T, T]:
    if end < start:
        raise ValueError("Invalid date range")
    return start, end


def validate_str_arg(name: str, value: Any, strip=True) -> str:
    if value is None:
        raise ValueError(f"Missing {name}")
    if not isinstance(value, str):
        raise TypeError(f"Invalid {name} type")
    if strip:
        value = value.strip()
    return value


def validate_int_arg(name: str, value: Any, default: Optional[int] = None) -> int:
    if value is None:
        if default is None:
            raise ValueError(f"Missing {name}")
        else:
            value = default
    if not isinstance(value, int):
        raise TypeError(f"Invalid {name} type")
    return value


def validate_finid_arg(name: str, value: Any) -> "FinId":
    from .structure import FinId

    if value is None:
        raise ValueError(f"Missing {name}")
    if isinstance(value, str):
        value = FinId.from_string(value)
    if not isinstance(value, FinId):
        raise TypeError(f"Invalid {name} type")
    return value


def validate_weekday_arg(name: str, value: Any) -> "Weekday":
    from .structure import Weekday

    if value is None:
        raise ValueError(f"Missing {name}")
    if isinstance(value, str):
        value = Weekday.from_string(value)
    if isinstance(value, int):
        value = Weekday(value)
    if isinstance(value, datetime.date):
        value = Weekday(value.weekday())
    if isinstance(value, datetime.datetime):
        value = Weekday(value.weekday())
    if not isinstance(value, Weekday):
        raise TypeError(f"Invalid {name} type")
    return value


def validate_path_arg(name: str, value: Any) -> Path:
    if value is None:
        raise ValueError(f"Missing {name}")
    if isinstance(value, str):
        value = Path(value)
    if not isinstance(value, Path):
        raise TypeError(f"Invalid {name} type")
    return value


def validate_subclass_arg(name: str, value: Any, baseclass: Type[T]) -> Type[T]:
    if value is None:
        raise ValueError(f"Missing {name}")
    if not issubclass(value, baseclass):
        raise TypeError(f"Invalid {name} type")
    return value


def validate_instance_arg(name: str, value: Any, baseclass: Type[T]) -> T:
    if value is None:
        raise ValueError(f"Missing {name}")
    if not isinstance(value, baseclass):
        raise TypeError(f"Invalid {name} type")
    return value



import csv
from pathlib import Path
from typing import Dict, Iterator, List, Optional, Tuple

from tradinghours.store.base import Cluster, Collection, Registry
from tradinghours.typing import StrOrPath
from tradinghours.validate import validate_path_arg


class FileCluster(Cluster):
    """Manages one page file with items for a collection"""

    DEFAULT_CACHE_SIZE = 500

    def __init__(self, location: StrOrPath, cache_size: Optional[int] = None):
        self._location = validate_path_arg("location", location)
        self._cached: List[str, Tuple] = []
        self._cache_size = cache_size or self.DEFAULT_CACHE_SIZE

    @property
    def location(self) -> Path:
        return self._location

    def truncate(self):
        with open(self.location, "a+", encoding="utf-8", newline="") as file:
            file.seek(0)
            file.truncate(0)

    def flush(self):
        with open(self.location, "a+", encoding="utf-8", newline="") as file:
            writer = csv.writer(file)
            for key, data in self._cached:
                row = [key, *data]
                writer.writerow(row)
            self._cached = []

    def load_all(self) -> Dict[str, Tuple]:
        keyed_items = {}
        next_key = 1
        with open(self.location, "r", encoding="utf-8", newline="") as file:
            for row in csv.reader(file):
                key = row[0] or str(next_key)
                next_key += 1
                data = row[1:]
                if key:
                    keyed_items[key] = data
        return keyed_items


class FileClusterRegistry(Registry[FileCluster]):
    """Holds a series of file clusters"""

    def __init__(self, folder: StrOrPath):
        self._folder = validate_path_arg("folder", folder)
        super().__init__()

    @property
    def folder(self) -> Path:
        return self._folder

    def create(self, slug: str) -> FileCluster:
        location = self.folder / f"{slug}.dat"
        return FileCluster(location)

    def discover(self) -> Iterator[str]:
        if self.folder.exists():
            for item in self.folder.iterdir():
                if item.is_file():
                    yield item.stem


class FileCollection(Collection):
    """Manages a collection of items in a store"""

    def __init__(self, folder: StrOrPath):
        self._folder = validate_path_arg("folder", folder)
        self._clusters = FileClusterRegistry(folder)

    @property
    def folder(self) -> Path:
        return self._folder

    @property
    def clusters(self) -> FileClusterRegistry:
        return self._clusters

    def touch(self):
        self.folder.mkdir(parents=True, exist_ok=True)

    def clear(self):
        for current in self._clusters:
            current.truncate()
        self._clusters = FileClusterRegistry(self.folder)


class FileCollectionRegistry(Registry[FileCollection]):
    """Holds a series of collections"""

    def __init__(self, root: StrOrPath):
        self._root = validate_path_arg("root", root)
        super().__init__()

    @property
    def root(self) -> Path:
        return self._root

    def create(self, slug: str) -> FileCollection:
        folder = self.root / slug
        collection = FileCollection(folder)
        collection.touch()
        return collection

    def discover(self) -> Iterator[str]:
        if self.root.exists():
            for item in self.root.iterdir():
                if item.is_dir():
                    yield item.name



import json
import unittest

from tradinghours.util import StrEncoder, slugify, snake_case, snake_dict


class TestSnakeCase(unittest.TestCase):
    def test_snake_case_with_dash(self):
        text = "hello-world"
        result = snake_case(text)
        self.assertEqual(result, "hello_world")

    def test_snake_case_no_dash(self):
        text = "helloWorld"
        result = snake_case(text)
        self.assertEqual(result, "hello_world")

    def test_snake_case_all_caps(self):
        text = "HELLO_WORLD"
        result = snake_case(text)
        self.assertEqual(result, "hello_world")

    def test_snake_case_none(self):
        text = None
        with self.assertRaises(ValueError):
            snake_case(text)

    def test_snake_case_int(self):
        text = 123
        with self.assertRaises(TypeError):
            snake_case(text)


class TestSnakeDict(unittest.TestCase):
    def test_snake_dict(self):
        data = {
            "First Name": "John",
            "Last Name": "Doe",
            "Age": 30,
        }
        result = snake_dict(data)
        expected_result = {
            "first_name": "John",
            "last_name": "Doe",
            "age": 30,
        }
        self.assertEqual(result, expected_result)

    def test_snake_dict_non_dict_input(self):
        data = "This is not a dictionary"
        with self.assertRaises(TypeError):
            snake_dict(data)


class TestSlugify(unittest.TestCase):
    def test_slugify_basic(self):
        name = "Hello World"
        result = slugify(name)
        self.assertEqual(result, "hello-world")

    def test_slugify_special_characters(self):
        name = "Hello!@#$%^World"
        result = slugify(name)
        self.assertEqual(result, "hello-world")

    def test_slugify_multiple_spaces(self):
        name = "Hello    World"
        result = slugify(name)
        self.assertEqual(result, "hello-world")

    def test_slugify_none(self):
        name = None
        with self.assertRaises(ValueError):
            slugify(name)

    def test_slugify_int(self):
        name = 123
        with self.assertRaises(TypeError):
            slugify(name)


class TestStrEncoder(unittest.TestCase):
    def test_str_encoder(self):
        encoder = StrEncoder
        data = {"name": "John", "age": 30}
        json_data = json.dumps(data, cls=encoder)
        self.assertEqual(json_data, '{"name": "John", "age": 30}')

    def test_str_encoder_non_serializable(self):
        encoder = StrEncoder
        data = {"name": "John", "age": 30, "non_serializable": set()}
        result = json.dumps(data, cls=encoder)
        expected_result = '{"name": "John", "age": 30, "non_serializable": "set()"}'
        self.assertEqual(result, expected_result)


if __name__ == "__main__":
    unittest.main()



import unittest

from tradinghours.market import Market


class EdgeCase(unittest.TestCase):
    def setUp(self):
        self.market = None

    def assertRangeSchedule(self, fin_id, start, end, expected):
        market = Market.get(fin_id)
        schedules = list(market.generate_schedules(start, end))
        phases = map(lambda s: s.phase_type, schedules)
        result = list(phases)
        self.assertEqual(result, expected)

    def assertDateSchedule(self, fin_id, date, expected):
        self.assertRangeSchedule(fin_id, date, date, expected)

    def assertRangeTimestamps(self, fin_id, start, end, expected):
        market = Market.get(fin_id)
        schedules = list(market.generate_schedules(start, end))
        phases = map(
            lambda s: (s.start.isoformat()[:19], s.end.isoformat()[:19]), schedules
        )
        result = list(phases)
        self.assertEqual(result, expected)

    def assertDateTimestamps(self, fin_id, date, expected):
        self.assertRangeTimestamps(fin_id, date, date, expected)


class TestCase001(EdgeCase):
    """

    Check there are correct schedules on a regularly open day, no holiday

    """

    def test_case(self):
        fin_id = "US.NYSE"
        date = "2023-11-15"
        expected = [
            "Pre-Trading Session",
            "Pre-Open",
            "Call Auction",
            "Primary Trading Session",
            "Pre-Close",
            "Post-Trading Session",
        ]
        self.assertDateSchedule(fin_id, date, expected)


class TestCase002(EdgeCase):
    """

    Check there are not schedules on a closed open day, no holiday

    """

    def test_case(self):
        fin_id = "US.NYSE"
        date = "2023-11-11"
        expected = []
        self.assertDateSchedule(fin_id, date, expected)


class TestCase003(EdgeCase):
    """

    Test there are correct schedules on an irregular schedule


    """

    def test_case(self):
        fin_id = "US.NYSE"
        date = "2023-11-24"
        expected = [
            "Pre-Trading Session",
            "Primary Trading Session",
            "Post-Trading Session",
        ]
        self.assertDateSchedule(fin_id, date, expected)


class TestCase004(EdgeCase):
    """

    Test there are correct schedules with schedule coming from the proceeding day, Regular Schedule (overnight)

    """

    def test_case(self):
        fin_id = "US.CME.EQUITY.USINDEX1"
        date = "2023-11-13"
        expected_phases = [
            "Primary Trading Session",
            "Pre-Open",
            "Primary Trading Session",
        ]
        self.assertDateSchedule(fin_id, date, expected_phases)
        expected_timestamps = [
            ("2023-11-12T17:00:00", "2023-11-13T16:00:00"),
            ("2023-11-13T16:45:00", "2023-11-13T17:00:00"),
            ("2023-11-13T17:00:00", "2023-11-14T16:00:00"),
        ]
        self.assertDateTimestamps(fin_id, date, expected_timestamps)


class TestCase005(EdgeCase):
    """

    Test there are correct schedules with schedule coming from the proceeding day, Irregular Schedule (overnight)

    """

    def test_case(self):
        fin_id = "US.CME.EQUITY.USINDEX1"
        date = "2023-11-23"
        expected_phases = [
            "Primary Trading Session",
            "Pre-Open",
            "Primary Trading Session",
        ]
        self.assertDateSchedule(fin_id, date, expected_phases)
        expected_timestamps = [
            ("2023-11-22T17:00:00", "2023-11-23T12:00:00"),
            ("2023-11-23T12:00:00", "2023-11-23T17:00:00"),
            ("2023-11-23T17:00:00", "2023-11-24T12:15:00"),
        ]
        self.assertDateTimestamps(fin_id, date, expected_timestamps)


class TestCase006(EdgeCase):
    """

    Test there are not schedules coming from the proceeding day when there is a holiday, but normally there would be an overnight schedule

    """

    def test_case(self):
        fin_id = "US.CME.EQUITY.USINDEX1"
        date = "2023-12-25"
        expected_phases = [
            "Pre-Open",
            "Primary Trading Session",
        ]
        self.assertDateSchedule(fin_id, date, expected_phases)
        expected_timestamps = [
            ("2023-12-25T16:00:00", "2023-12-25T17:00:00"),
            ("2023-12-25T17:00:00", "2023-12-26T16:00:00"),
        ]
        self.assertDateTimestamps(fin_id, date, expected_timestamps)


class TestCase007(EdgeCase):
    """

    Test there are correct schedules on a working Weekend (If Saturday is set
    as Regular in the holidays table, but the regular schedule is normally
    M-F, ignore the day of week.)

    """

    def test_case(self):
        fin_id = "CN.CIBM"
        date = "2020-01-19"
        expected = [
            "Primary Trading Session",
            "Intermission",
            "Primary Trading Session",
        ]
        self.assertDateSchedule(fin_id, date, expected)


class TestCase008(EdgeCase):
    """

    Test the correct schedule for the day of the week is returned for schedule
    with different hours on different days of the week

    """

    def test_case_thu(self):
        fin_id = "US.CBOE.VIX"
        date = "2020-10-15"
        # TODO: review this order
        expected_phases = [
            "Primary Trading Session",
            "Trading-at-Last",
            "Primary Trading Session",
            "Post-Trading Session",
            "Pre-Open",
            "Primary Trading Session",
            "Trading-at-Last",
        ]
        self.assertDateSchedule(fin_id, date, expected_phases)
        expected_timestamps = [
            ("2020-10-14T17:00:00", "2020-10-15T08:30:00"),
            ("2020-10-14T17:00:00", "2020-10-15T15:00:00"),
            ("2020-10-15T08:30:00", "2020-10-15T15:00:00"),
            ("2020-10-15T15:00:00", "2020-10-15T16:00:00"),
            ("2020-10-15T16:45:00", "2020-10-15T17:00:00"),
            ("2020-10-15T17:00:00", "2020-10-16T08:30:00"),
            ("2020-10-15T17:00:00", "2020-10-16T15:00:00"),
        ]
        self.assertDateTimestamps(fin_id, date, expected_timestamps)

    def test_case_fri(self):
        fin_id = "US.CBOE.VIX"
        date = "2020-10-16"
        # TODO: review this order
        expected_phases = [
            "Primary Trading Session",
            "Trading-at-Last",
            "Primary Trading Session",
            "Post-Trading Session",
        ]
        self.assertDateSchedule(fin_id, date, expected_phases)
        expected_timestamps = [
            ("2020-10-15T17:00:00", "2020-10-16T08:30:00"),
            ("2020-10-15T17:00:00", "2020-10-16T15:00:00"),
            ("2020-10-16T08:30:00", "2020-10-16T15:00:00"),
            ("2020-10-16T15:00:00", "2020-10-16T16:00:00"),
        ]
        self.assertDateTimestamps(fin_id, date, expected_timestamps)


class TestCase009(EdgeCase):
    """

    Test there are correct schedules on irregular day when the irregular schedule
    does have a schedule for the current day of the week

    """

    def test_case_sun(self):
        fin_id = "US.CME.AGRI.DAIRY1"
        date = "2022-01-16"
        expected_phases = [
            "Pre-Open",
        ]
        self.assertDateSchedule(fin_id, date, expected_phases)
        expected_timestamps = [
            ("2022-01-16T16:00:00", "2022-01-17T17:00:00"),
        ]
        self.assertDateTimestamps(fin_id, date, expected_timestamps)

    def test_case_mon(self):
        fin_id = "US.CME.AGRI.DAIRY1"
        date = "2022-01-17"
        expected_phases = [
            "Pre-Open",
            "Primary Trading Session",
        ]
        self.assertDateSchedule(fin_id, date, expected_phases)
        expected_timestamps = [
            ("2022-01-16T16:00:00", "2022-01-17T17:00:00"),
            ("2022-01-17T17:00:00", "2022-01-18T16:00:00"),
        ]
        self.assertDateTimestamps(fin_id, date, expected_timestamps)

    def test_case_reg_sun(self):
        fin_id = "US.CME.AGRI.DAIRY1"
        date = "2022-01-09"
        expected_phases = [
            "Pre-Open",
            "Primary Trading Session",
        ]
        self.assertDateSchedule(fin_id, date, expected_phases)
        expected_timestamps = [
            ("2022-01-09T16:00:00", "2022-01-09T17:00:00"),
            ("2022-01-09T17:00:00", "2022-01-10T16:00:00"),
        ]
        self.assertDateTimestamps(fin_id, date, expected_timestamps)

    def test_case_reg_mon(self):
        fin_id = "US.CME.AGRI.DAIRY1"
        date = "2022-01-10"
        expected_phases = [
            "Primary Trading Session",
            "Pre-Open",
            "Primary Trading Session",
        ]
        self.assertDateSchedule(fin_id, date, expected_phases)
        expected_timestamps = [
            ("2022-01-09T17:00:00", "2022-01-10T16:00:00"),
            ("2022-01-10T16:45:00", "2022-01-10T17:00:00"),
            ("2022-01-10T17:00:00", "2022-01-11T16:00:00"),
        ]
        self.assertDateTimestamps(fin_id, date, expected_timestamps)


class TestCase010(EdgeCase):
    """

    Test Seasonality cases

    """

    def test_case_season(self):
        fin_id = "US.BTEC.ACTIVES.US"
        date = "2023-03-09"
        expected_phases = [
            "Primary Trading Session",
            "Primary Trading Session",
        ]
        self.assertDateSchedule(fin_id, date, expected_phases)
        expected_timestamps = [
            ("2023-03-08T18:30:00", "2023-03-09T17:30:00"),
            ("2023-03-09T18:30:00", "2023-03-10T17:30:00"),
        ]
        self.assertDateTimestamps(fin_id, date, expected_timestamps)

    def test_overnight(self):
        fin_id = "US.BTEC.ACTIVES.US"
        date = "2023-11-12"
        expected = [
            ("2023-11-12T18:30:00", "2023-11-13T17:30:00"),
        ]
        self.assertDateTimestamps(fin_id, date, expected)


class TestCase011(EdgeCase):
    """

    Testing holiday with offset

    """

    def test_case_single(self):
        fin_id = "CN.SGE"
        date = "2023-01-02"
        expected = []
        self.assertDateTimestamps(fin_id, date, expected)

    def test_case_multi(self):
        fin_id = "CN.SGE"
        start_date = "2023-01-01"
        end_date = "2023-01-03"
        expected = [
            ("2023-01-03T09:00:00", "2023-01-03T15:30:00"),
            ("2023-01-03T15:00:00", "2023-01-03T15:30:00"),
            ("2023-01-03T15:30:00", "2023-01-03T15:45:00"),
            ("2023-01-03T15:31:00", "2023-01-03T15:40:00"),
            ("2023-01-03T15:40:00", "2023-01-03T15:40:00"),
            ("2023-01-03T19:45:00", "2023-01-03T20:00:00"),
            ("2023-01-03T19:50:00", "2023-01-03T19:59:00"),
            ("2023-01-03T20:00:00", "2023-01-04T02:30:00"),
        ]
        self.assertRangeTimestamps(fin_id, start_date, end_date, expected)


class TestCase012(EdgeCase):
    """

    Test whether you can follow or not a permanently closed market

    """

    def setUp(self):
        self.mic = "XBUE"
        self.old_finid = "AR.BCBA"
        self.new_finid = "AR.BYMA"

    def test_follow_auto_mic(self):
        market = Market.get(self.mic)
        result = str(market.fin_id)
        expected = self.new_finid
        self.assertEqual(result, expected)

    def test_original_auto_mic(self):
        market = Market.get(self.mic, follow=False)
        result = str(market.fin_id)
        expected = self.old_finid
        self.assertEqual(result, expected)

    def test_follow_auto_finid(self):
        market = Market.get(self.old_finid)
        result = str(market.fin_id)
        expected = self.new_finid
        self.assertEqual(result, expected)

    def test_original_auto_finid(self):
        market = Market.get(self.old_finid, follow=False)
        result = str(market.fin_id)
        expected = self.old_finid
        self.assertEqual(result, expected)

    def test_follow_by_mic(self):
        market = Market.get_by_mic(self.mic)
        result = str(market.fin_id)
        expected = self.new_finid
        self.assertEqual(result, expected)

    def test_original_by_mic(self):
        market = Market.get_by_mic(self.mic, follow=False)
        result = str(market.fin_id)
        expected = self.old_finid
        self.assertEqual(result, expected)

    def test_follow_by_finid(self):
        market = Market.get_by_finid(self.old_finid)
        result = str(market.fin_id)
        expected = self.new_finid
        self.assertEqual(result, expected)

    def test_original_by_finid(self):
        market = Market.get_by_finid(self.old_finid, follow=False)
        result = str(market.fin_id)
        expected = self.old_finid
        self.assertEqual(result, expected)


class TestCase013(EdgeCase):
    """

    Test whether MIC case is ignored

    """

    def test_upper_finid(self):
        market = Market.get_by_finid("AR.BYMA")
        result = str(market.fin_id)
        expected = "AR.BYMA"
        self.assertEqual(result, expected)

    def test_lower_finid(self):
        market = Market.get_by_finid("ar.byma")
        result = str(market.fin_id)
        expected = "AR.BYMA"
        self.assertEqual(result, expected)

    def test_mixed_finid(self):
        market = Market.get_by_finid("aR.ByMa")
        result = str(market.fin_id)
        expected = "AR.BYMA"
        self.assertEqual(result, expected)

    def test_upper_mic(self):
        market = Market.get_by_mic("XBUE")
        result = str(market.fin_id)
        expected = "AR.BYMA"
        self.assertEqual(result, expected)

    def test_lower_mic(self):
        market = Market.get_by_mic("xbue")
        result = str(market.fin_id)
        expected = "AR.BYMA"
        self.assertEqual(result, expected)

    def test_mixed_mic(self):
        market = Market.get_by_mic("xBuE")
        result = str(market.fin_id)
        expected = "AR.BYMA"
        self.assertEqual(result, expected)



import datetime
from dataclasses import dataclass, field
from typing import List, Tuple

import pytz

from .typing import WeekdaySpec
from .validate import validate_weekday_arg


@dataclass
class OlsonTimezone:
    country: str
    detail: str

    @property
    def tzinfo(self):
        return pytz.timezone(str(self))

    def localize(self, dt: datetime.datetime) -> datetime.datetime:
        return self.tzinfo.localize(dt)

    def __str__(self):
        return f"{self.country}/{self.detail}"

    @classmethod
    def from_string(cls, input_string: str):
        country, _, detail = input_string.partition("/")
        return cls(country, detail)


@dataclass
class Weekday:
    WEEKDAYS = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]
    ordinal: int

    def next(self) -> "Weekday":
        if self.ordinal == 6:
            return Weekday(0)
        return Weekday(self.ordinal + 1)

    def previous(self) -> "Weekday":
        if self.ordinal == 0:
            return Weekday(6)
        return Weekday(self.ordinal - 1)

    @classmethod
    def from_string(cls, input_string):
        input_string = input_string.title()
        if input_string in cls.WEEKDAYS:
            return cls(cls.WEEKDAYS.index(input_string))
        else:
            raise ValueError("Invalid weekday string")

    def __str__(self):
        return self.WEEKDAYS[self.ordinal]


@dataclass
class WeekdayPeriod:
    start_day: Weekday
    end_day: Weekday

    @property
    def weekdays(self) -> List:
        captured = []
        current_day = self.start_day.previous()
        finished = False
        while not finished:
            current_day = current_day.next()
            captured.append(current_day)
            if current_day == self.end_day:
                finished = True
        return captured

    def matches(self, other: WeekdaySpec) -> bool:
        other = validate_weekday_arg("other", other)
        return other in self.weekdays

    @classmethod
    def from_string(cls, input_string):
        if "-" in input_string:
            start, end = input_string.split("-")
            start_day = Weekday.from_string(start)
            end_day = Weekday.from_string(end)
            return cls(start_day, end_day)
        else:
            only_day = Weekday.from_string(input_string)
            return cls(only_day, only_day)

    def __str__(self):
        if self.start_day == self.end_day:
            return str(self.start_day)
        else:
            return f"{str(self.start_day)}-{str(self.end_day)}"


@dataclass
class WeekdaySet:
    periods: Tuple[WeekdayPeriod]

    def matches(self, other: WeekdaySpec) -> bool:
        other = validate_weekday_arg("other", other)
        for current in self.periods:
            if current.matches(other):
                return True
        return False

    @classmethod
    def from_string(cls, input_string):
        all_period_str = input_string.split(",")
        all_period = []
        for period_str in all_period_str:
            period = WeekdayPeriod.from_string(period_str)
            all_period.append(period)
        return cls(tuple(all_period))

    def __str__(self):
        periods = [str(p) for p in self.periods]
        return ",".join(periods)


@dataclass
class FinId:
    country: str
    acronym: str
    extra: list = field(default_factory=list)

    def __post_init__(self):
        self.country = self.country.upper()
        self.acronym = self.acronym.upper()
        self.extra = [e.upper() for e in self.extra]

    @classmethod
    def from_string(cls, input_string):
        segments = input_string.split(".")
        if len(segments) < 2:
            raise ValueError("Invalid FinID string")
        return cls(segments[0], segments[1], segments[2:])

    def __str__(self):
        segments = [self.country, self.acronym] + self.extra
        return ".".join(segments)


@dataclass
class Mic:
    code: str

    def __post_init__(self):
        self.code = self.code.upper()

    @classmethod
    def from_string(cls, input_string):
        return cls(input_string)

    def __str__(self):
        return self.code



import argparse
import time
import traceback
from contextlib import contextmanager
from textwrap import wrap
from threading import Thread

from tradinghours import __version__
from tradinghours.catalog import default_catalog
from tradinghours.currency import Currency
from tradinghours.exceptions import TradingHoursError
from tradinghours.market import Market
from tradinghours.remote import default_data_manager

EXIT_CODE_EXPECTED_ERROR = 1
EXIT_CODE_UNKNOWN_ERROR = 2


def print_help(text):
    lines = wrap(text, initial_indent="  ", subsequent_indent="  ")
    print("\n  --")
    print("\n".join(lines))
    print()


@contextmanager
def timed_action(message: str):
    start = time.time()
    print(f"{message}...", end="", flush=True)

    done = False

    def print_dots():
        while not done:
            print(".", end="", flush=True)
            time.sleep(0.5)

    thread = Thread(target=print_dots)
    thread.daemon = True
    thread.start()

    yield start

    elapsed = time.time() - start
    done = True
    thread.join()
    print(f" ({elapsed:.3f}s)")


def create_parser():
    parser = argparse.ArgumentParser(description="TradingHours API Client")

    # Create a subparser for the subcommands
    subparsers = parser.add_subparsers(dest="command", help="Available subcommands")
    subparsers.required = True

    # "status" subcommand
    status_parser = subparsers.add_parser("status", help="Get status")
    status_parser.add_argument(
        "--extended", action="store_true", help="Show more information"
    )

    # "import" subcommand
    import_parser = subparsers.add_parser("import", help="Import data")
    import_parser.add_argument("--force", action="store_true", help="Force the import")

    return parser


def run_status(args):
    with timed_action("Collecting timestamps"):
        remote_timestamp = default_data_manager.remote_timestamp
        local_timestamp = default_data_manager.local_timestamp
    print("TradingHours Data Status:")
    print("  Remote Timestamp:  ", remote_timestamp.ctime())
    print("  Local Timestamp:   ", local_timestamp and local_timestamp.ctime())
    print()
    if args.extended:
        if local_timestamp:
            with timed_action("Reading local data"):
                all_currencies = list(Currency.list_all())
                all_markets = list(Market.list_all())
            print("Extended Information:")
            print("  Currencies count:  ", len(all_currencies))
            print("  Markets count:     ", len(all_markets))
        else:
            print("No local data to show extended information")


def run_import(args):
    if args.force or default_data_manager.needs_download:
        with timed_action("Downloading"):
            default_data_manager.download()
        with timed_action("Ingesting"):
            default_catalog.ingest_all()
    else:
        print("Local data is up-to-date.")


def main():
    try:
        # Main console entrypoint
        parser = create_parser()
        args = parser.parse_args()
        if args.command == "status":
            run_status(args)
        elif args.command == "import":
            run_import(args)

    # Handle generic errors gracefully
    except Exception as error:
        # TradingHours errors with help messages are simpler
        if isinstance(error, TradingHoursError) and error.help_message:
            print("ERROR:", error.detail)
            print_help(error.help_message)
            exit(EXIT_CODE_EXPECTED_ERROR)

        # Other errors will generate a traceback dump
        error_message = f"ERROR: {error}"
        print(error_message)

        try:
            # Try saving extra information to local file
            traceback_info = traceback.format_exc()
            version_message = f"\nVERSION: {__version__}"
            with open("debug.txt", "w") as debug_file:
                debug_file.write(error_message)
                debug_file.write(version_message)
                debug_file.write("\n\nTraceback:\n")
                debug_file.write(traceback_info)
            print_help(
                "Details about this error were saved to debug.txt file. You can "
                "submit it to the support team for further investigation. Feel "
                "free also to submit the file in a new GitHub issue.",
            )
        except Exception as error:
            print("Failed saving debug information.", error)
        finally:
            exit(EXIT_CODE_UNKNOWN_ERROR)


if __name__ == "__main__":
    main()



import datetime

from .base import BaseObject, DateField, IntegerField, StringField
from .exceptions import MissingDefinitionError
from .validate import validate_int_arg, validate_str_arg


class SeasonDefinition(BaseObject):
    """Seasonality information"""

    season = StringField()
    """Season entry or code"""

    year = IntegerField()
    """Year when this definition is value"""

    date = DateField()
    """Especific date for this year"""

    @classmethod
    def get(cls, season: str, year: int, catalog=None) -> "SeasonDefinition":
        season = validate_str_arg("season", season)
        year = validate_int_arg("year", year)
        catalog = cls.get_catalog(catalog)
        for _, definition in catalog.list(cls):
            if definition.season.lower() == season.lower() and definition.year == year:
                return definition
        raise MissingDefinitionError()

    @classmethod
    def get_date(cls, season: str, year: int, catalog=None) -> datetime.date:
        definition = cls.get(season, year)
        return definition.date



import datetime
import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch
from urllib.request import HTTPError

from tradinghours.exceptions import ClientError, TokenError
from tradinghours.remote import Client, DataManager


class TestClientResponse(unittest.TestCase):
    def setUp(self):
        self.client = Client(token="test_token", base_url="http://example.com")
        self.mock_urlopen = patch("tradinghours.remote.urlopen").start()

    def tearDown(self):
        patch.stopall()

    def test_urlopen_successful(self):
        self.mock_urlopen.return_value = "foobar"

        with self.client.get_response("/test") as response:
            self.assertEqual(response, "foobar")

    def test_urlopen_token_error(self):
        self.mock_urlopen.side_effect = HTTPError(
            "url", 401, "Unauthorized", None, None
        )

        with self.assertRaises(TokenError):
            with self.client.get_response("/test"):
                pass

    def test_urlopen_client_error(self):
        self.mock_urlopen.side_effect = HTTPError(
            "url", 500, "Server Error", None, None
        )

        with self.assertRaises(ClientError):
            with self.client.get_response("/test"):
                pass


class ClientTestCase(unittest.TestCase):
    @property
    def client(self):
        if not hasattr(self, "_client"):
            self._client = Client(token="test_token", base_url="http://example.com")
        return self._client

    def patchResponse(self, content):
        content = content.encode("utf-8")
        mock_response = MagicMock()
        mock_response.__enter__.return_value.read.side_effect = [content, b""]
        mock_response.__enter__.return_value.__exit__.return_value = False
        patcher = patch(
            "tradinghours.remote.Client.get_response", return_value=mock_response
        )
        return patcher

    def patchResponseError(self, exception):
        mock_response = MagicMock()
        mock_response.__enter__.side_effect = exception
        patcher = patch(
            "tradinghours.remote.Client.get_response", return_value=mock_response
        )
        return patcher

    def patchResponseFile(self, file_path):
        with open(file_path, "rb") as file:
            content = file.read()

        mock_response = MagicMock()
        mock_response.__enter__.return_value.read.side_effect = [content, b""]
        mock_response.__enter__.return_value.__exit__.return_value = False
        patcher = patch(
            "tradinghours.remote.Client.get_response", return_value=mock_response
        )
        return patcher


class TestClientJson(ClientTestCase):
    def test_get_json_successful(self):
        with self.patchResponse('{"key": "value"}'):
            data = self.client.get_json("/test")
            self.assertEqual(data, {"key": "value"})

    def test_get_json_token_error(self):
        with self.patchResponseError(TokenError("Token is missing or invalid")):
            with self.assertRaises(TokenError):
                self.client.get_json("/test")

    def test_get_json_client_error(self):
        with self.patchResponseError(ClientError("Error getting server response")):
            with self.assertRaises(ClientError):
                self.client.get_json("/test")


class TestClientDownload(ClientTestCase):
    def test_download_temporary_successful(self):
        with self.patchResponse('{"key": "value"}'):
            with self.client.download_temporary("/test") as temp_file:
                content = temp_file.read().decode("utf-8")
                self.assertEqual(content, '{"key": "value"}')

    def test_download_temporary_token_error(self):
        with self.patchResponseError(TokenError("Token is missing or invalid")):
            with self.assertRaises(TokenError):
                with self.client.download_temporary("/test"):
                    pass

    def test_download_temporary_client_error(self):
        with self.patchResponseError(ClientError("Error getting server response")):
            with self.assertRaises(ClientError):
                with self.client.download_temporary("/test"):
                    pass


# Define a custom function to use as a side effect
def custom_validate_instance_arg(arg_name, arg_value, expected_type):
    return arg_value


class TestDataManager(ClientTestCase):
    @property
    def manager(self):
        if not hasattr(self, "_manager"):
            self._manager = DataManager(
                client=MagicMock(spec=Client),
                root=Path(__file__).parent / "fixtures" / "remote",
            )
        return self._manager

    @patch(
        "tradinghours.remote.validate_instance_arg",
        side_effect=custom_validate_instance_arg,
    )
    @patch("tradinghours.remote.default_client.get_json")
    def test_remote_timestamp(self, mock_get_json, mock_validate_instance_arg):
        timestamp = "2023-10-27T12:00:00"
        mock_get_json.return_value = {"last_updated": timestamp}
        expected_datetime = datetime.datetime.fromisoformat(timestamp)

        self.assertEqual(self.manager.remote_timestamp, expected_datetime)

    def test_download(self):
        with self.patchResponseFile("sample_data.zip") as patcher:
            self.manager.download()

        patcher.stop()



from typing import List

from .base import (
    BaseObject,
    BooleanField,
    DateField,
    OlsonTimezoneField,
    StringField,
    WeekdaySetField,
)
from .typing import StrOrDate
from .validate import validate_date_arg, validate_range_args, validate_str_arg


class Currency(BaseObject):
    """A currency supported by TradingHours."""

    code = StringField()
    """3-letter code of the currency (ISO 4217)."""

    name = StringField()
    """English name of the currency."""

    country_code = StringField()
    """2-letter country code for the currency's country."""

    central_bank = StringField()
    """Name of the central bank for the currency."""

    financial_capital = StringField()
    """City where the central bank is located."""

    financial_capital_timezone = OlsonTimezoneField()
    """Timezone Olson timezone identifier format."""

    weekend = WeekdaySetField()
    """Weekend definition. Most markets are Sat-Sun."""

    def list_holidays(
        self, start: StrOrDate, end: StrOrDate, catalog=None
    ) -> "CurrencyHoliday":
        start, end = validate_range_args(
            validate_date_arg("start", start),
            validate_date_arg("end", end),
        )
        catalog = self.get_catalog(catalog)
        holidays = list(
            catalog.filter(
                CurrencyHoliday,
                cluster=self.currency_code,
                key_from=start.isoformat(),
                key_to=end.isoformat(),
            )
        )
        return holidays

    @classmethod
    def list_all(cls, catalog=None) -> List["Currency"]:
        catalog = cls.get_catalog(catalog)
        return list(catalog.list_all(Currency))

    @classmethod
    def get(cls, code: str, catalog=None) -> "Currency":
        code = validate_str_arg("code", code)
        catalog = cls.get_catalog(catalog)
        return catalog.get(cls, code)


class CurrencyHoliday(BaseObject):
    """Holiday for an specific currency"""

    currency_code = StringField()
    """3-letter code of the currency (ISO 4217)."""

    date = DateField()
    """Shows the date of the holiday for the currency."""

    name = StringField()
    """Describes the name of the holiday."""

    has_settlement = BooleanField()
    """Whether the market has settlement for the holiday."""

    observed = BooleanField()
    """Whether the holiday is observed."""

    memo = StringField()
    """A description or additional details about the holiday, if applicable."""



from typing import Dict, Generator, Optional, Tuple, Type, TypeVar

from tradinghours.validate import validate_str_arg

from .base import BaseObject
from .currency import Currency, CurrencyHoliday
from .market import Market, MarketHoliday, MicMapping
from .remote import default_data_manager
from .schedule import Schedule
from .season import SeasonDefinition
from .store.engine import Store, default_store
from .store.file import FileCollection
from .store.source import SourceFile
from .typing import StrOrPath

B = TypeVar("B", bound=BaseObject)


class DeclaredFile(SourceFile[B]):
    """Well known source file"""

    known_files: Dict[str, Type["DeclaredFile"]] = {}

    name: str = None
    model: Type[B] = None

    def __init__(self, root: StrOrPath):
        super().__init__(root, self.name, self.model)

    def __init_subclass__(cls):
        super().__init_subclass__()
        if cls.name is None:
            raise ValueError("name must be defined")
        if cls.model is None:
            raise ValueError("model must be defined")
        cls.known_files[cls.name] = cls

    def pre_ingest(self, store: "Store"):
        pass

    def resolve_cluster(self, item: Type[B]) -> Optional[str]:
        return None

    def resolve_key(self, item: Type[B]) -> Optional[str]:
        return None

    def ingest(self, store: "Store"):
        self.pre_ingest(store)
        for current in self.load_iter():
            collection = self.name
            cluster = self.resolve_cluster(current)
            key = self.resolve_key(current)
            key = str(key).upper() if key else key
            data = current.to_tuple()
            store.store_tuple(data, collection, cluster=cluster, key=key)


class SeasonDefinitionFile(DeclaredFile):
    name = "season-definitions"
    model = SeasonDefinition


class CurrencyFile(DeclaredFile[Currency]):
    name = "currencies"
    model = Currency

    def resolve_key(self, item: Currency) -> Optional[str]:
        return item.code


class CurrencyHolidayFile(DeclaredFile[CurrencyHoliday]):
    name = "currency_holidays"
    model = CurrencyHoliday

    def resolve_cluster(self, item: CurrencyHoliday) -> Optional[str]:
        return item.currency_code

    def resolve_key(self, item: CurrencyHoliday) -> Optional[str]:
        return item.date


class MarketFile(DeclaredFile[Market]):
    name = "markets"
    model = Market

    def resolve_cluster(self, item: MarketHoliday) -> Optional[str]:
        return item.fin_id.country

    def resolve_key(self, item: MarketHoliday) -> Optional[str]:
        return str(item.fin_id)


class MarketHolidayFile(DeclaredFile[MarketHoliday]):
    name = "holidays"
    model = MarketHoliday

    def resolve_cluster(self, item: MarketHoliday) -> Optional[str]:
        return str(item.fin_id)

    def resolve_key(self, item: MarketHoliday) -> Optional[str]:
        return item.date


class MicMappingFile(DeclaredFile[MicMapping]):
    name = "mic-mapping"
    model = MicMapping

    def resolve_key(self, item: MicMapping) -> Optional[str]:
        return item.mic


class ScheduleFile(DeclaredFile[Schedule]):
    name = "schedules"
    model = Schedule

    def resolve_cluster(self, item: MarketHoliday) -> Optional[str]:
        return str(item.fin_id)

    def pre_ingest(self, store: "Store"):
        store.clear_collection(self.name)


class Catalog:
    """Gives you access to an underlying data store

    TODO: maybe DeclaredFile and Store.ingest all should be here

    """

    def __init__(self, store: Store):
        self._store = store

    @property
    def store(self) -> Store:
        return self._store

    def ingest_all(self):
        for _, declared_class in DeclaredFile.known_files.items():
            source = declared_class(default_data_manager.csv_dir)
            source.ingest(self.store)
        self.store.flush()

    def find_model_collection(self, model: Type[BaseObject]) -> FileCollection:
        for name, declared in DeclaredFile.known_files.items():
            if declared.model is model:
                return self.store.collections.get(name)
        return None

    def list_all(self, model: Type[B]) -> Generator[B, None, None]:
        collection = self.find_model_collection(model)
        for cluster in collection.clusters:
            cluster_data = cluster.load_all()
            for _, current in cluster_data.items():
                yield model.from_tuple(current)

    def list(
        self, model: Type[B], cluster: Optional[str] = None
    ) -> Generator[Tuple[str, B], None, None]:
        collection = self.find_model_collection(model)
        cluster_name = cluster or "default"
        cluster_obj = collection.clusters.get(cluster_name)
        cluster_data = cluster_obj.load_all()
        for current_key, data in cluster_data.items():
            yield current_key, model.from_tuple(data)
        return None

    def get(
        self, model: Type[B], key: str, cluster: Optional[str] = None
    ) -> Optional[B]:
        key = validate_str_arg("key", key).upper()
        collection = self.find_model_collection(model)
        cluster_name = cluster or "default"
        cluster = collection.clusters.get(cluster_name)
        cluster_data = cluster.load_all()
        for current_key, data in cluster_data.items():
            if current_key == key:
                return model.from_tuple(data)
        return None

    def filter(
        self,
        model: Type[B],
        key_start: str,
        key_end: str,
        cluster: Optional[str] = None,
    ) -> Generator[B, None, None]:
        key_start = validate_str_arg("key_start", key_start).upper()
        key_end = validate_str_arg("key_end", key_end).upper()
        collection = self.find_model_collection(model)
        cluster_name = cluster or "default"
        cluster = collection.clusters.get(cluster_name)
        cluster_data = cluster.load_all()
        for current_key, data in cluster_data.items():
            if current_key >= key_start and current_key <= key_end:
                yield model.from_tuple(data)
        return None


default_catalog = Catalog(default_store)


if __name__ == "__main__":
    from datetime import date
    from time import time

    print("\nDownloading...")
    start = time()
    default_data_manager.download()
    elapsed = time() - start
    print("Elapsed seconds", elapsed)

    print("\nImporting...")
    start = time()
    default_catalog.ingest_all()
    elapsed = time() - start
    print("Elapsed seconds", elapsed)

    print("\nLoading Market...")
    start = time()
    us_market: Market = default_catalog.get(Market, "US.IEX", "us")
    elapsed = time() - start
    print("Elapsed seconds", elapsed)

    print("\nListing Holidays...")
    start = time()
    loaded = list(us_market.list_holidays("2023-06-01", "2023-12-31"))
    elapsed = time() - start
    print("Elapsed seconds", elapsed, len(loaded))

    print("\nListing Currencies...")
    start = time()
    loaded = list(Currency.list_all())
    elapsed = time() - start
    print("Elapsed seconds", elapsed, len(loaded))

    print("\nListing Schedules...")
    start = time()
    loaded = list(Schedule.list_all())
    elapsed = time() - start
    print("Elapsed seconds", elapsed, len(loaded))

    print("\nGenerating Schedules...")
    start = time()
    for concrete in us_market.generate_schedules(date(2023, 9, 1), date(2023, 9, 30)):
        print(concrete)
    elapsed = time() - start
    print("Elapsed seconds", elapsed, len(loaded))

    print("\nDone")
